@@ -10,6 +10,7 @@ use std::io;
 use std::ops::{Deref, DerefMut};
 use std::os::unix::io::{FromRawFd, IntoRawFd};
 use std::result::Result;
+use std::sync::atomic::Ordering;
 use std::sync::{Arc, Mutex, MutexGuard, RwLock, RwLockReadGuard, RwLockWriteGuard};
 
 use virtio_queue::{Error as VirtQueError, Queue, QueueT};
@@ -45,7 +46,7 @@ pub trait VringT<M: GuestAddressSpace>:
     /// Add an used descriptor into the used queue.
     fn add_used(&self, desc_index: u16, len: u32) -> Result<(), VirtQueError>;
 
-    /// Notify the vhost-user master that used descriptors have been put into the used queue.
+    /// Notify the vhost-user frontend that used descriptors have been put into the used queue.
     fn signal_used_queue(&self) -> io::Result<()>;
 
     /// Enable event notification for queue.
@@ -74,6 +75,12 @@ pub trait VringT<M: GuestAddressSpace>:
     /// Set queue next avail head.
     fn set_queue_next_avail(&self, base: u16);
 
+    /// Set queue next used head.
+    fn set_queue_next_used(&self, idx: u16);
+
+    /// Get queue next used head index from the guest memory.
+    fn queue_used_idx(&self) -> Result<u16, VirtQueError>;
+
     /// Set configured queue size.
     fn set_queue_size(&self, num: u16);
 
@@ -138,7 +145,7 @@ impl<M: GuestAddressSpace> VringState<M> {
             .add_used(self.mem.memory().deref(), desc_index, len)
     }
 
-    /// Notify the vhost-user master that used descriptors have been put into the used queue.
+    /// Notify the vhost-user frontend that used descriptors have been put into the used queue.
     pub fn signal_used_queue(&self) -> io::Result<()> {
         if let Some(call) = self.call.as_ref() {
             call.write(1)
@@ -192,6 +199,18 @@ impl<M: GuestAddressSpace> VringState<M> {
         self.queue.set_next_avail(base);
     }
 
+    /// Set queue next used head.
+    fn set_queue_next_used(&mut self, idx: u16) {
+        self.queue.set_next_used(idx);
+    }
+
+    /// Get queue next used head index from the guest memory.
+    fn queue_used_idx(&self) -> Result<u16, VirtQueError> {
+        self.queue
+            .used_idx(self.mem.memory().deref(), Ordering::Relaxed)
+            .map(|idx| idx.0)
+    }
+
     /// Set configured queue size.
     fn set_queue_size(&mut self, num: u16) {
         self.queue.set_size(num);
@@ -326,6 +345,14 @@ impl<M: 'static + GuestAddressSpace> VringT<M> for VringMutex<M> {
         self.lock().set_queue_next_avail(base)
     }
 
+    fn set_queue_next_used(&self, idx: u16) {
+        self.lock().set_queue_next_used(idx)
+    }
+
+    fn queue_used_idx(&self) -> Result<u16, VirtQueError> {
+        self.lock().queue_used_idx()
+    }
+
     fn set_queue_size(&self, num: u16) {
         self.lock().set_queue_size(num);
     }
@@ -433,6 +460,14 @@ impl<M: 'static + GuestAddressSpace> VringT<M> for VringRwLock<M> {
         self.write_lock().set_queue_next_avail(base)
     }
 
+    fn set_queue_next_used(&self, idx: u16) {
+        self.write_lock().set_queue_next_used(idx)
+    }
+
+    fn queue_used_idx(&self) -> Result<u16, VirtQueError> {
+        self.get_ref().queue_used_idx()
+    }
+
     fn set_queue_size(&self, num: u16) {
         self.write_lock().set_queue_size(num);
     }
